import asyncio
import json
import logging
from urllib.parse import urljoin

from aiohttp import ClientResponse, ClientSession

import config
from gestion_sports.gestion_sports_payload_builder import GestionSportsPayloadBuilder
from models import BookingFilter, Club, User

DATE_FORMAT = "%d/%m/%Y"

TIME_FORMAT = "%H:%M"

LOGGER = logging.getLogger(__name__)
POST_HEADERS = config.get_post_headers("gestion-sports")


class GestionSportsConnector:
    """
    Handle the specific booking requests to Gestion-Sports
    """

    def __init__(self, session: ClientSession, url: str):
        LOGGER.info("Initializing connection to GestionSports API")
        self.url = url
        self.session = session
        self.payload_builder = GestionSportsPayloadBuilder()

    @property
    def landing_url(self) -> str:
        """
        Get the URL to the landing page of Gestion-Sports

        :return: the URL to the landing page
        """
        return urljoin(self.url, "/connexion.php?")

    @property
    def login_url(self) -> str:
        """
        Get the URL to the connection login of Gestion-Sports

        :return: the URL to the login page
        """
        return urljoin(self.url, "/connexion.php?")

    @property
    def booking_url(self) -> str:
        """
        Get the URL to the booking page of Gestion-Sports

        :return: the URL to the booking page
        """
        return urljoin(self.url, "/membre/reservation.html?")

    async def land(self) -> ClientResponse:
        """
        Perform the request to the landing page in order to get the cookie PHPSESSIONID

        :return: the response from the landing page
        """
        LOGGER.info("Connecting to GestionSports API")
        async with self.session.get(self.landing_url) as response:
            await response.text()
            return response

    async def login(self, user: User, club: Club) -> ClientResponse:
        """
        Perform the request to the log in the user

        :return: the response from the login
        """
        payload = (
            self.payload_builder.login(user.login)
            .password(user.password)
            .club_id(club.id)
            .build_login_payload()
        )

        async with self.session.post(
            self.login_url, data=payload, headers=POST_HEADERS
        ) as response:
            await response.text()
            return response

    async def book(self, booking_filter: BookingFilter, club: Club) -> int | None:
        """
        Perform a request for each court at the same time to increase the chances to get a booking.
        The gestion-sports backend does not allow several bookings at the same time
        so there is no need to make each request one after the other

        :param booking_filter: the booking information
        :param club: the club where to book the court
        :return: the booked court, or None if no court was booked
        """
        # use asyncio to request a booking on every court
        # the gestion-sports backend is able to book only one court for a user
        bookings = await asyncio.gather(
            *[
                self.book_one_court(booking_filter, court_id)
                for court_id in club.courts_ids
            ],
            return_exceptions=True,
        )

        return await self.get_booked_court(bookings)

    @staticmethod
    async def get_booked_court(bookings):
        for court, is_booked in bookings:
            if is_booked:
                return court
        return None

    async def book_one_court(
        self, booking_filter: BookingFilter, court_id: int
    ) -> tuple[int, bool]:
        payload = (
            self.payload_builder.date(booking_filter.date.date().strftime(DATE_FORMAT))
            .time(booking_filter.date.time().strftime(TIME_FORMAT))
            .sport_id(booking_filter.sport_id)
            .court_id(court_id)
            .build_booking_payload()
        )
        return court_id, await self.is_court_booked(payload)

    async def is_court_booked(self, payload: str) -> bool:
        async with self.session.post(
            self.booking_url, data=payload, headers=POST_HEADERS
        ) as response:
            return self.is_response_status_ok(await response.text())

    @staticmethod
    def is_response_status_ok(response: str) -> bool:
        formatted_result = response.removeprefix('"').removesuffix('"')
        result_json = json.loads(formatted_result)
        return result_json["status"] == "ok"
